package ru.dbotthepony.kstarbound.world.entities.behavior

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonPrimitive
import org.classdump.luna.ByteString
import org.classdump.luna.Table
import org.classdump.luna.Userdata
import org.classdump.luna.impl.ImmutableTable
import ru.dbotthepony.kstarbound.defs.actor.behavior.NodeParameter
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.luaFunction
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.util.valueOf
import java.util.EnumMap
import java.util.HashSet
import kotlin.collections.HashMap

/**
 * Blackboard (Greenboard, if you will) for tracking inputs and outputs of function nodes in behavior tree.
 *
 * While this is obviously flawed, and inputs/outputs should _generally_ be stored inside nodes themselves
 * (since we create/clone behavior tree each time we make AI actor), this class is exposed to Lua scripts,
 * so we must keep this class
 *
 * On side note, this architecture also employs global inputs/outputs of functions, hence you can't make
 * an actual functional tree out of behavior nodes (because variable flow is global, and not local, so you
 * can't have, per say, function "calculate sin out of x" because both input "x" and "sin result" are
 * defined as global variables).
 * Bravo, Chucklefish.
 */
class Blackboard(val lua: LuaEnvironment) : Userdata<Blackboard>() {
	private val board = EnumMap<NodeParameterType, HashMap<String, Any>>(NodeParameterType::class.java)
	private val input = EnumMap<NodeParameterType, HashMap<String, ArrayList<Pair<String, Table>>>>(NodeParameterType::class.java)
	private val parameters = HashMap<Any, Table>()
	// key -> list of Lua tables and their indices
	private val vectorNumberInput = HashMap<String, HashSet<Pair<Any, Table>>>()
	private var ephemeral = HashSet<Pair<NodeParameterType, String>>()

	init {
		for (v in NodeParameterType.entries) {
			board[v] = HashMap()
			input[v] = HashMap()
		}
	}

	operator fun set(type: NodeParameterType, key: String, value: Any?) {
		if (value == null) {
			board[type]!!.remove(key)
		} else {
			board[type]!![key] = value
		}

		val input = input[type]!![key]

		if (input != null) {
			for ((index, table) in input) {
				table[index] = value
			}
		}

		// dumb special case for setting number outputs to vec2 inputs
		if (type == NodeParameterType.NUMBER) {
			val mappings = vectorNumberInput[key]

			if (mappings != null) {
				for ((index, table) in mappings) {
					table[index] = value
				}
			}
		}
	}

	operator fun get(type: NodeParameterType, key: String): Any? {
		return board[type]!![key]
	}

	fun parameters(parameters: Map<String, NodeParameter>, nodeID: Any): Table {
		val get = this.parameters[nodeID]

		if (get != null)
			return get

		val table = lua.newTable()

		for ((name, parameter) in parameters.entries) {
			if (parameter.value.key != null) {
				val typeInput = input[parameter.type]!!.computeIfAbsent(parameter.value.key) { ArrayList() }
				typeInput.add(name to table)
				table[name] = this[parameter.type, parameter.value.key]
			} else {
				val value = parameter.value.value ?: JsonNull.INSTANCE

				if (value.isJsonNull)
					continue

				// dumb special case for allowing a vec2 of blackboard number keys
				if (parameter.type == NodeParameterType.VEC2) {
					if (value !is JsonArray)
						throw IllegalArgumentException("Expected '$name' to be VEC2 array, got $value")

					val vector = lua.newTable(value.size(), 0)

					for ((i, element) in value.withIndex()) {
						if (element is JsonPrimitive && element.isString) {
							vectorNumberInput.computeIfAbsent(element.asString) { HashSet() }.add(i + 1L to vector)
							vector[i + 1L] = this[NodeParameterType.NUMBER, element.asString]
						} else {
							vector[i + 1L] = lua.from(element)
						}
					}

					table[name] = vector
				} else {
					table[name] = lua.from(value)
				}
			}
		}

		this.parameters[nodeID] = table
		return table
	}

	fun setOutput(node: ActionNode, output: Table) {
		for ((tableKey, out) in node.outputs) {
			if (out.key != null) {
				this[out.type, out.key] = output[tableKey]

				if (out.ephemeral) {
					ephemeral.add(out.type to out.key)
				}
			}
		}
	}

	fun takeEphemerals(): Set<Pair<NodeParameterType, String>> {
		val ephemeral = ephemeral
		this.ephemeral = HashSet()
		return ephemeral
	}

	fun clearEphemerals(ephemerals: Collection<Pair<NodeParameterType, String>>) {
		for ((type, key) in ephemerals) {
			this[type, key] = null
		}
	}

	override fun getMetatable(): Table {
		return Companion.metatable
	}

	override fun setMetatable(mt: Table?): Table {
		throw UnsupportedOperationException()
	}

	override fun getUserValue(): Blackboard {
		return this
	}

	override fun setUserValue(value: Blackboard?): Blackboard {
		throw UnsupportedOperationException()
	}

	companion object {
		private val metatable: ImmutableTable
		private fun __index() = metatable

		init {
			val builder = ImmutableTable.Builder()
				.add("get", luaFunction { self: Blackboard, type: ByteString, key: ByteString ->
					returnBuffer.setTo(self[NodeParameterType.entries.valueOf(type.decode()), key.decode()])
				})
				.add("set", luaFunction { self: Blackboard, type: ByteString, key: ByteString, value: Any? ->
					self[NodeParameterType.entries.valueOf(type.decode()), key.decode()] = value
				})

			for (type in NodeParameterType.entries) {
				builder.add("get${type.funcName}", luaFunction { self: Blackboard, key: ByteString ->
					returnBuffer.setTo(self[type, key.decode()])
				})

				builder.add("set${type.funcName}", luaFunction { self: Blackboard, key: ByteString, value: Any? ->
					self[type, key.decode()] = value
				})
			}

			builder.add("__index", luaFunction { _: Any?, index: Any -> returnBuffer.setTo(__index()[index]) })
			metatable = builder.build()
		}
	}
}
