package ru.dbotthepony.kstarbound.world.physics

import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import java.util.*

enum class CollisionType(
	override val jsonName: String,

	/**
	 * Consider as solid when calculating collisions
	 */
	val isSolidCollision: Boolean,

	/**
	 * Consider as solid if we are standing on it when calculating collisions
	 *
	 * We deviate from original engine here, we consider [DYNAMIC] as solid floor,
	 * where original engine does not (despite closing trapdoors should make complete sense to be used as floor;
	 * [DYNAMIC] not being considered floor is the reason why crewmates jump over vertical ship hatches on BYOS ships and similar)
	 */
	val isFloorCollision: Boolean,

	/**
	 * Consider as solid when pathfinding
	 */
	val isSolidNonDynamic: Boolean
) : IStringSerializable {
	/**
	 * Not loaded, block collisions by default
	 */
	NULL("Null", true, true, true),

	/**
	 * Air
	 */
	NONE("None", false, false, false),

	/**
	 * Platforms themselves and including stairs made of platforms
	 */
	PLATFORM("Platform", false, true, false),

	/**
	 * Collision formed dynamically and can be changed at any time, e.g. by doors
	 *
	 * Locked doors use [BLOCK] instead of [DYNAMIC] though, because their "open" condition can't be realistically predicted
	 */
	DYNAMIC("Dynamic", true, true, false),

	/**
	 * Solid collision; by default used by bottom and top world borders
	 */
	SLIPPERY("Slippery", true, true, true),

	/**
	 * Solid collision by any means
	 */
	BLOCK("Block", true, true, true);

	fun maxOf(other: CollisionType): CollisionType {
		if (this === NULL || other === NULL)
			return NULL
		else if (this > other)
			return this
		else
			return other
	}

	companion object {
		val SOLID: Set<CollisionType> = Collections.unmodifiableSet(EnumSet.copyOf(entries.filter { it.isSolidCollision }.toSet()))
	}
}
